//---------------------------------------------------------------------------------------------------------------------------------------------------
// Last Change: 2023-01-25      Name: import_.cpp
//---------------------------------------------------------------------------------------------------------------------------------------------------
// This file contains the language-independent implementation of the plug-in IMPORT_.DLL. All texts and resources that are language-dependent are
// located in an additional IMPORT.DLL, of which sources can be found in the subdirectories \E (for English) and \D (for German).
//---------------------------------------------------------------------------------------------------------------------------------------------------

#define MK_PRODUCT_REGKEY L"IMPORT"

//---------------------------------------------------------------------------------------------------------------------------------------------------

#include  "windows.h"
#include  "windowsx.h"
#include  "stdlib.h"
#include  "stdio.h"
#include  "math.h"

#ifdef _MSC_VER
#pragma warning( disable:4458 ) // declaration hides class member
#endif

#include  "gdiplus.h"

#ifdef _MSC_VER
#pragma warning( default:4458 ) // declaration hides class member
#endif

#include  "d2d1.h"
#include  "d2d1_1.h"
#include  "dwrite.h"
#include  "wincodec.h"

#include  "x:\mkcad6\mki_lib6\mki_lib6.h"
#include  "x:\mkcad6\mki_lib6\CompileDate.h"
#include  "x:\mkcad6\mki_lib6\Dialog.h"

//---------------------------------------------------------------------------------------------------------------------------------------------------
// Language-dependent texts in IMPORT.DLL

constexpr LPSTR TXT_ID_HEADER = "Malz++Kassner TXT 6.00";

MKI_INTERFACE_IMPORT LPWSTR
  eStartUpText[],
  eDialogText[];

//---------------------------------------------------------------------------------------------------------------------------------------------------

class CGlobalData
{

  public:

  CGlobalData( void )
  {
    ZeroMemory( this, sizeof( *this ) );
  }

  HINSTANCE m_hInstDLL;     // Instance handle of the main DLL
  HINSTANCE m_hLanguage;    // Instance handle of the language DLL
  HINSTANCE m_hGlobalInst;  // Instance handle of the serving application
  HWND      m_hGlobalWnd;   // Main window handle of the serving application
  
  MKI_BITMAP_PTR m_pBitmap; // Handle of command icon bitmap

  MKI_POINT m_cOrigin;  // Current origin
  __int32   m_nError;   // Current error status ( 0 = OK )
};

CGlobalData gGlobalData;


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure reads a single coordinate pair from the current file. If it finds the last line of the file (indicated by the MKI_DB_END data
// block identifier), it returns false without setting m_nError to a non-zero value.
bool
PlugInReadCoordinate( MKI_POINT_REF f_rPoint )
{
  MKI_DUMMYSTRW szText1, 
                szText2;
  __int16       nDummy;

  if( gGlobalData.m_nError )
    return( false );

  if( MKI_DialogProgressIsPending() )
  {
    MKI_ConvertNumberString( szText2, MKI_FileReadCurrentLine() );
    MKI_PrintW( szText1, eDialogText[3], szText2 );
    MKI_ConvertBytesString( szText2, MKI_FileReadCurrentSize() );
    MKI_DialogUpdateProgress( szText1, szText2, MKI_FileReadCurrentSize(), MKI_FileReadTotalSize() );

    if( MKI_DialogIsCancelled() )
    {
      gGlobalData.m_nError = 999;
      return( false );
    }
  }

  MKI_FileReadInt16( nDummy );
  if( nDummy == MKI_DB_END )
    return( false );

  MKI_FileReadComma(); 
  MKI_FileReadDouble( f_rPoint.x );
  MKI_FileReadComma(); 
  MKI_FileReadDouble( f_rPoint.y );
  MKI_FileReadSemi();

  if( MKI_FileReadError() )
  {
    gGlobalData.m_nError = 2;
    return( false );
  }

  return( true );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure performs the import. It reads all coordinate pairs from the import file and insert a marking for each coordinate pair.
// In addition, it initializes and display a progress indicator window to inform the user about the current progress and to allow him to
// cancel the import. Since the final file size is already known in advance, the progress indicator does include a percent bar.
void
PlugInImport( HANDLE f_hFileHandle, const LPCWSTR f_pszFileName )
{
  MKI_FILENAMEW szFileName;
  MKI_DUMMYSTRW szDummyStr;
  MKI_DUMMYSTRA szTextA;
  MKI_SYSTEMDEF cSystemDef;
  MKI_POINT     cPoint;

  gGlobalData.m_nError = 0;

  if( MKI_SystemGetCurrent( cSystemDef ) >= 0 )
  {
    gGlobalData.m_cOrigin = cSystemDef.m_cSystemOrigin;
  }
  else
  {
    gGlobalData.m_cOrigin.SetZero();
  }

  if( !MKI_FileReadInitDisk( f_hFileHandle, false ) )
    return;

  // Skip first line
  if( !MKI_FileReadLine( szTextA, sizeof( szTextA ) ) )
  {
    gGlobalData.m_nError = 3;
    MKI_FileReadExit();
    return;
  }

  MKI_FileBaseName( f_pszFileName, szFileName );
  MKI_PrintW( szDummyStr, eDialogText[2], szFileName );
  MKI_DialogShowProgress( eDialogText[0], szDummyStr, true );

  if( PlugInReadCoordinate( cPoint ) )
  {
    __int32 nCount = 0;

    do
    {
      if( nCount == 0 )
        MKI_ObjectOpen( MKI_OBJ_MARK );

      MKI_ObjectAddPoint( MKI_DB_POINT_MARK, cPoint.x + gGlobalData.m_cOrigin.x, cPoint.y + gGlobalData.m_cOrigin.y );
      nCount++;

      if( nCount >= MKI_MaxEntityPoints() )
      {
        if( !MKI_ObjectFastInsert( nullptr ) )
        {
          MKI_PrintW( szDummyStr, L"nCount: %d\nMKI_MaxEntityPoints(): %d", nCount, MKI_MaxEntityPoints() );
          MKI_DialogMessage( szDummyStr, eDialogText[0], MKI_MESSAGE_OK );

          gGlobalData.m_nError = 998;
          MKI_FileReadExit();
          MKI_DialogHideProgress();
          return;
        }
        nCount = 0;
      }
    }
    while( PlugInReadCoordinate( cPoint ) );

    if( nCount > 0 )
    {
      if( !MKI_ObjectFastInsert( nullptr ) )
        gGlobalData.m_nError = 998;
    }

    MKI_FileReadSemi();
  }

  MKI_FileReadExit();
  MKI_DialogHideProgress();
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This DLL entry procedure must exist in any DLL to be used in Win32. Since our DLL does all necessary initialization in its MKI_PlugInInit()
// procedure, this procedure is quite empty.
BOOL WINAPI
DllMain( HINSTANCE f_hInstance, DWORD f_dwReason, LPVOID f_pDummy )
{
  f_pDummy; // Touch formal parameter to avoid C4100

  switch( f_dwReason )
  {
    case DLL_PROCESS_ATTACH:
      gGlobalData.m_hInstDLL = f_hInstance;
      break;

    case DLL_PROCESS_DETACH:
      gGlobalData.m_hInstDLL = nullptr;
      break;
  }
  return( true );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when the plug-in is loaded by the serving application. Its main tasks are:
// - Checking whether it is compatible with the given f_nInterfaceVersion
// - Checking whether it is licensed to the given serial number (optional)
// - Storing of the serving application's instance and main windows handle for further use
// - Loading of the language-dependent library
// - Filling in the plug-in ID structure of which address is passed in f_pPlugInID
// - Loading of options from the registry database
// - Loading profiles
// - Allocating any static memory required
MKI_INTERFACE_EXPORT __int32
MKI_PlugInInit( const LPCSTR f_pszSerialNumber, HINSTANCE f_hMainInst, HWND f_hMainWnd,
                __int32 f_nInterfaceVersion, MKI_PLUGIN_ID_REF f_rPlugInID )
{
  f_pszSerialNumber;  // Touch formal parameter to avoid C4100

  MKI_FILENAMEW szFileName;

  if( f_nInterfaceVersion > MKI_INTERFACE_VERSION )
    return( MKI_INIT_OLD_PLUGIN );
  else if( f_nInterfaceVersion < MKI_INTERFACE_VERSION )
    return( MKI_INIT_OLD_APPLICATION );

  gGlobalData.m_hGlobalInst = f_hMainInst;
  gGlobalData.m_hGlobalWnd  = f_hMainWnd;

  gGlobalData.m_hLanguage = LoadLibraryW( MK_PRODUCT_REGKEY L".DLL" );
  MKI_FileApplicationPath( MK_PRODUCT_REGKEY L".PNG", szFileName );
  gGlobalData.m_pBitmap = MKI_BitmapLoadImage( szFileName );

  // Fill the plug-in information data block
  f_rPlugInID.m_nOwnerID    = MKI_DB_OWNER_MK;
  f_rPlugInID.m_nPlugInID   = 11000;
  f_rPlugInID.m_nPlugInCTRL = MKI_PLUGINCTRL_STARTER | MKI_PLUGINCTRL_PRO | MKI_PLUGINCTRL_TEST | MKI_PLUGINCTRL_DEMO | MKI_PLUGINCTRL_READONLY;
  f_rPlugInID.m_pContainer  = nullptr;

  f_rPlugInID.m_cPlugInProc.m_pfnInputPointInitProc = (MKI_INPUTPOINTINIT_PROC) nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputPointMoveProc = (MKI_INPUTPOINTMOVE_PROC) nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputPointExitProc = (MKI_INPUTPOINTEXIT_PROC) nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputDisplayProc   = (MKI_INPUTDISPLAY_PROC)   nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputParameterProc = (MKI_INPUTPARAMETER_PROC) nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputCancelProc    = (MKI_INPUTCANCEL_PROC)    nullptr;
  f_rPlugInID.m_cPlugInProc.m_pfnInputFinishProc    = (MKI_INPUTFINISH_PROC)    nullptr;

  f_rPlugInID.m_cPlugInData.m_nType = MKI_PLUGINMENU_IMPORT;

  f_rPlugInID.m_cPlugInData.m_cInputData.m_nCommandMode = MKI_COMMAND_DIRECT;
  f_rPlugInID.m_cPlugInData.m_cMenuData.m_pszMenuEntry   = eStartUpText[1];
  f_rPlugInID.m_cPlugInData.m_cMenuData.m_pszDescription = eStartUpText[2];

  f_rPlugInID.m_cPlugInData.m_cIconData.m_pIconBitmap  = gGlobalData.m_pBitmap;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconXOffset = MKI_ICON_COMMAND_FRAME;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconYOffset = MKI_ICON_COMMAND_FRAME;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconXSize   = MKI_ICON_COMMAND_SIZE;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconYSize   = MKI_ICON_COMMAND_SIZE;
  f_rPlugInID.m_cPlugInData.m_cIconData.m_nIconMode    = 0;

  f_rPlugInID.m_pCommandData  = nullptr;
  f_rPlugInID.m_pButtonData   = nullptr;
  f_rPlugInID.m_pExtendedData = nullptr;

  f_rPlugInID.m_pszFileType      = eStartUpText[3];
  f_rPlugInID.m_pszFileExtension = eStartUpText[4];
  f_rPlugInID.m_pszFileDefault   = eStartUpText[5];

  return( MKI_INIT_OK );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when the plug-in is removed by the serving application. Its main tasks are:
// - Checking whether anything is to be saved. If so, it should display a message information the user
//   about it and allowing him to save those changes.
// - Freeing of all statically allocated memory.
// If this procedure returns false, the serving application will not be able to terminate. So please, do only return false if shutting down the
// plug-in now would severely damage or destroy user data.
MKI_INTERFACE_EXPORT bool
MKI_PlugInExit( bool f_fSaveSettings )
{
  f_fSaveSettings;  // Touch formal parameter to avoid C4100

  MKI_BitmapFreeImage( gGlobalData.m_pBitmap );

  return( true );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when the user wants to load a non-MKD drawing using the File>Open command. For an import filter, its task is to check
// whether he knows this type of file, and to load it, if possible.
MKI_INTERFACE_EXPORT __int32
MKI_PlugInImport( const LPCWSTR f_pszFileName, __int32 f_nFilterMode, void* f_pData )
{
  f_pData;  // Touch formal parameter to avoid C4100

  HANDLE        hFileHandle;
  MKI_DUMMYSTRW szDummyStr;
  MKI_DUMMYSTRA szTextA;
  bool          fFound = false;

  SetCursor( LoadCursor( nullptr, IDC_WAIT ) );

//
// Step 1: Check whether this file is a valid file containing coordinate pairs.
//

  if( MKI_FileOpen( &hFileHandle, f_pszFileName ) )
  {
    if( MKI_FileUnbufferedRead( hFileHandle, szTextA, MKI_LENGTH_FILE_ID + 1 ) == MKI_LENGTH_FILE_ID + 1 )
    {
      szTextA[MKI_LENGTH_FILE_ID] = '\0';
      if( MKI_EqualA( szTextA, TXT_ID_HEADER ) )
        fFound = true;
    }
    MKI_FileClose( hFileHandle );
  }
  if( !fFound )
    return( MKI_IMPORT_UNKNOWN );

//
// Step 2: Try to import the coordinates stored in the file.
//

  if( MKI_FileOpen( &hFileHandle, f_pszFileName ) )
  {
    if( f_nFilterMode & MKI_FILTER_MODE_MERGE )
      MKI_UndoInitProcess();
    else
      MKI_DrawingNewFile( MKI_FILEMODE_USERSECTIONS, false );

    PlugInImport( hFileHandle, f_pszFileName );

    MKI_FileClose( hFileHandle );
    switch( gGlobalData.m_nError )
    {
      case 999:
        MKI_DialogMessage( eDialogText[4], eDialogText[0], MKI_MESSAGE_OK );
        goto _error;

      case 998:
        MKI_DialogMessage( eDialogText[5], eDialogText[0], MKI_MESSAGE_OK );
        goto _error;

      case 0:
        break;

      default:
        MKI_PrintW( szDummyStr, eDialogText[6], gGlobalData.m_nError, MKI_FileReadCurrentLine(), (int) MKI_FileReadCurrentSize() );
        MKI_DialogMessage( szDummyStr, eDialogText[0], MKI_MESSAGE_OK );
        goto _error;
    }

    if( f_nFilterMode & MKI_FILTER_MODE_MERGE )
    {
      MKI_UndoFinishProcess( MKI_CHANGECODE_ADD );
      MKI_UndoSetPrevious();
    }

//  MKI_UndoUpdateLinks();      <- Would only be necessary if blocks or instances had been modified!
    MKI_DrawWindowAll();

    return( MKI_IMPORT_OK );
  }
  return( MKI_IMPORT_ERROR );

_error:
  if( f_nFilterMode & MKI_FILTER_MODE_MERGE )
    MKI_UndoCancelProcess();
  else
    MKI_DrawingNewFile( MKI_FILEMODE_USERSECTIONS, false );

  return( MKI_IMPORT_ERROR );
}


//---------------------------------------------------------------------------------------------------------------------------------------------------
// This procedure is called when a plug-in's command is chosen by the user. For an import filter, its main tasks are:
// - Parameter editing
// - Help file display
MKI_INTERFACE_EXPORT __int32
MKI_PlugInCommand( __int32 f_nCommandID, __int32 f_nExecMode, void* f_pData )
{
  f_nCommandID; // Touch formal parameter to avoid C4100
  f_pData;      // Touch formal parameter to avoid C4100

  // Check which type of action is requested
  switch( f_nExecMode )
  {
    // Display the corresponding help topic
    case MKI_EXECMODE_HELP:
      MKI_DialogHelpTopic( gGlobalData.m_hGlobalWnd, MK_PRODUCT_REGKEY, 0 );
      return( true );

    // Edit the options
    case MKI_EXECMODE_USER:
      {
        MKI_PLUGIN_CONTAINER  cData;

        MKI_CopyW( cData.m_szRegistry, MK_PRODUCT_REGKEY );
        MKI_GetBuildDate( __DATE__, cData.m_szBuildDate );
        cData.m_pszCaption   = eDialogText[0];
        cData.m_pszVendor    = nullptr;
        cData.m_pszCopyright = eStartUpText[0];
        cData.m_nUserMode    = MKI_PLUGINMODE_FREE;

        MKI_DialogPlugInAbout( GetActiveWindow(), cData, nullptr );
      }
      return( true );

    // Execute the command
    case MKI_EXECMODE_SYSTEM:
      // Nothing to do here...
      break;

    // Read drawing-dependent data
    case MKI_EXECMODE_GET_PROFILE:
      // Nothing to do here...
      break;

    // Write drawing-dependent data
    case MKI_EXECMODE_SET_PROFILE:
      // Nothing to do here...
      break;

    // Undo selected by user
    case MKI_EXECMODE_UNDO:
      // Nothing to do here...
      break;

    // Redo selected by user
    case MKI_EXECMODE_REDO:
      // Nothing to do here...
      break;

    // Clear undo data
    case MKI_EXECMODE_CLEAR_UNDO:
      // Nothing to do here...
      break;
  }
  return( false );
}
